@@ -1,4 +1,4 @@
-# Copyright (c) 2009 Evgeni Golov <evgeni.golov@uni-duesseldorf.de>
+# Copyright (c) 2009-2014 Evgeni Golov <evgeni@golov.de>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -31,19 +31,30 @@ from twisted.internet import defer
 from twisted.internet import reactor
 
 import datetime
-from bleyhelpers import *
+import logging
+import re
+
+import bleyhelpers
 from postfix import PostfixPolicy
 
 from time import sleep
 
+import ipaddr
+
+logger = logging.getLogger('bley')
+regexp_type = type(re.compile(''))
+
+
 class BleyPolicy(PostfixPolicy):
     '''Implementation of intelligent greylisting based on `PostfixPolicy`'''
 
     db = None
     dbc = None
+    required_params = ['sender', 'recipient', 'client_address',
+                       'client_name', 'helo_name']
 
     @defer.inlineCallbacks
-    def check_policy (self):
+    def check_policy(self):
         '''Check the incoming mail based on our policy and tell Postfix
         about our decision.
 
@@ -72,38 +83,51 @@ class BleyPolicy(PostfixPolicy):
             except:
                 self.safe_reconnect()
 
-        check_results = {'DNSWL': 0, 'DNSBL': 0, 'HELO': 0, 'DYN': 0, 'DB': -1, 'SPF': 0, 'S_EQ_R': 0 }
+        check_results = {'DNSWL': 0, 'DNSBL': 0, 'HELO': 0, 'DYN': 0, 'DB': -1,
+                         'SPF': 0, 'S_EQ_R': 0, 'WHITELISTED': 0, 'CACHE': 0}
         action = 'DUNNO'
+        self.params['now'] = datetime.datetime.now()
         postfix_params = self.params
 
-        # Strip everything after a + in the localpart, usefull for mailinglists etc
+        # Strip everything after a + in the localpart,
+        # usefull for mailinglists etc
         if postfix_params['sender'].find('+') != -1:
-            postfix_params['sender'] = postfix_params['sender'][:postfix_params['sender'].find('+')]+postfix_params['sender'][postfix_params['sender'].find('@'):]
+            postfix_params['sender'] = postfix_params['sender'][:postfix_params['sender'].find('+')] + postfix_params['sender'][postfix_params['sender'].find('@'):]
         if postfix_params['recipient'].find('+') != -1:
-            postfix_params['recipient'] = postfix_params['recipient'][:postfix_params['recipient'].find('+')]+postfix_params['recipient'][postfix_params['recipient'].find('@'):]
+            postfix_params['recipient'] = postfix_params['recipient'][:postfix_params['recipient'].find('+')] + postfix_params['recipient'][postfix_params['recipient'].find('@'):]
 
         if postfix_params['client_address'] in self.factory.bad_cache.keys():
-            delta = datetime.datetime.now()-self.factory.bad_cache[postfix_params['client_address']]
-            if delta < datetime.timedelta(0,60,0):
+            delta = datetime.datetime.now() - self.factory.bad_cache[postfix_params['client_address']]
+            if delta < datetime.timedelta(0, self.factory.settings.cache_valid, 0):
                 action = 'DEFER_IF_PERMIT %s (cached result)' % self.factory.settings.reject_msg
+                check_results['CACHE'] = 1
                 if self.factory.settings.verbose:
-                    self.factory.settings.logger('decided CACHED action=%s, checks: %s, postfix: %s\n' % (action, check_results, postfix_params))
+                    logger.info('decided CACHED action=%s, checks: %s, postfix: %s' %
+                                (action, check_results, postfix_params))
                 else:
-                    self.factory.settings.logger('decided CACHED action=%s, from=%s, to=%s\n' % (action, postfix_params['sender'], postfix_params['recipient']))
+                    logger.info('decided CACHED action=%s, from=%s, to=%s' %
+                                (action, postfix_params['sender'],
+                                 postfix_params['recipient']))
                 self.send_action(action)
+                self.factory.log_action(postfix_params, action, check_results)
                 return
             else:
                 del self.factory.bad_cache[postfix_params['client_address']]
 
         if postfix_params['client_address'] in self.factory.good_cache.keys():
-            delta = datetime.datetime.now()-self.factory.good_cache[postfix_params['client_address']]
-            if delta < datetime.timedelta(0,60,0):
+            delta = datetime.datetime.now() - self.factory.good_cache[postfix_params['client_address']]
+            if delta < datetime.timedelta(0, self.factory.settings.cache_valid, 0):
                 action = 'DUNNO'
+                check_results['CACHE'] = 1
                 if self.factory.settings.verbose:
-                    self.factory.settings.logger('decided CACHED action=%s, checks: %s, postfix: %s\n' % (action, check_results, postfix_params))
+                    logger.info('decided CACHED action=%s, checks: %s, postfix: %s' %
+                                (action, check_results, postfix_params))
                 else:
-                    self.factory.settings.logger('decided CACHED action=%s, from=%s, to=%s\n' % (action, postfix_params['sender'], postfix_params['recipient']))
+                    logger.info('decided CACHED action=%s, from=%s, to=%s' %
+                                (action, postfix_params['sender'],
+                                 postfix_params['recipient']))
                 self.send_action(action)
+                self.factory.log_action(postfix_params, action, check_results)
                 return
             else:
                 del self.factory.good_cache[postfix_params['client_address']]
@@ -113,31 +137,41 @@ class BleyPolicy(PostfixPolicy):
         #  0 : regular host, not in black, not in white, let it go
         #  1 : regular host, but in white, let it go, dont check EHLO
         #  2 : regular host, but in black, lets grey for now
-        if postfix_params['recipient'].lower().startswith('postmaster'):
+        if self.check_whitelist(postfix_params['recipient'].lower(),
+                                self.factory.settings.whitelist_recipients):
             action = 'DUNNO'
-        elif status == -1: # not found in local db...
+            check_results['WHITELISTED'] = 1
+        elif self.check_whitelist(postfix_params['client_name'].lower(),
+                                  self.factory.settings.whitelist_clients):
+            action = 'DUNNO'
+            check_results['WHITELISTED'] = 1
+        elif self.check_whitelist_ip(postfix_params['client_address'].lower(),
+                                     self.factory.settings.whitelist_clients_ip):
+            action = 'DUNNO'
+            check_results['WHITELISTED'] = 1
+        elif status == -1:  # not found in local db...
             check_results['DNSWL'] = yield self.check_dnswls(postfix_params['client_address'], self.factory.settings.dnswl_threshold)
             if check_results['DNSWL'] >= self.factory.settings.dnswl_threshold:
                 new_status = 1
             else:
                 check_results['DNSBL'] = yield self.check_dnsbls(postfix_params['client_address'], self.factory.settings.dnsbl_threshold)
-                check_results['HELO'] = check_helo(postfix_params)
-                check_results['DYN'] = check_dyn_host(postfix_params['client_name'])
+                check_results['HELO'] = bleyhelpers.check_helo(postfix_params)
+                check_results['DYN'] = bleyhelpers.check_dyn_host(postfix_params['client_name'])
                 # check_sender_eq_recipient:
-                if postfix_params['sender']==postfix_params['recipient']:
+                if postfix_params['sender'] == postfix_params['recipient']:
                     check_results['S_EQ_R'] = 1
-                if check_results['DNSBL'] < self.factory.settings.dnsbl_threshold and check_results['HELO']+check_results['DYN']+check_results['S_EQ_R'] < self.factory.settings.rfc_threshold:
-                    check_results['SPF'] = check_spf(postfix_params)
+                if self.factory.settings.use_spf and check_results['DNSBL'] < self.factory.settings.dnsbl_threshold and check_results['HELO'] + check_results['DYN'] + check_results['S_EQ_R'] < self.factory.settings.rfc_threshold:
+                    check_results['SPF'] = bleyhelpers.check_spf(postfix_params, self.factory.settings.use_spf_guess)
                 else:
                     check_results['SPF'] = 0
-                if check_results['DNSBL'] >= self.factory.settings.dnsbl_threshold or check_results['HELO']+check_results['DYN']+check_results['SPF']+check_results['S_EQ_R'] >= self.factory.settings.rfc_threshold:
+                if check_results['DNSBL'] >= self.factory.settings.dnsbl_threshold or check_results['HELO'] + check_results['DYN'] + check_results['SPF'] + check_results['S_EQ_R'] >= self.factory.settings.rfc_threshold:
                     new_status = 2
                     action = 'DEFER_IF_PERMIT %s' % self.factory.settings.reject_msg
                     self.factory.bad_cache[postfix_params['client_address']] = datetime.datetime.now()
                 else:
                     new_status = 0
                     self.factory.good_cache[postfix_params['client_address']] = datetime.datetime.now()
-            query = "INSERT INTO bley_status (ip, status, last_action, sender, recipient) VALUES(%(client_address)s, %(new_status)s, NOW(), %(sender)s, %(recipient)s)"
+            query = "INSERT INTO bley_status (ip, status, last_action, sender, recipient) VALUES(%(client_address)s, %(new_status)s, %(now)s, %(sender)s, %(recipient)s)"
             postfix_params['new_status'] = new_status
             try:
                 self.safe_execute(query, postfix_params)
@@ -145,12 +179,12 @@ class BleyPolicy(PostfixPolicy):
                 # the other thread already commited while we checked, ignore
                 pass
 
-        elif status[0] >= 2: # found to be greyed
+        elif status[0] >= 2:  # found to be greyed
             check_results['DB'] = status[0]
-            delta = datetime.datetime.now()-status[1]
-            if delta > self.factory.settings.greylist_period+status[2]*self.factory.settings.greylist_penalty or delta > self.factory.settings.greylist_max:
+            delta = datetime.datetime.now() - status[1]
+            if delta > self.factory.settings.greylist_period + status[2] * self.factory.settings.greylist_penalty or delta > self.factory.settings.greylist_max:
                 action = 'DUNNO'
-                query = "UPDATE bley_status SET status=0, last_action=NOW() WHERE ip=%(client_address)s AND sender=%(sender)s AND recipient=%(recipient)s"
+                query = "UPDATE bley_status SET status=0, last_action=%(now)s WHERE ip=%(client_address)s AND sender=%(sender)s AND recipient=%(recipient)s"
                 self.factory.good_cache[postfix_params['client_address']] = datetime.datetime.now()
             else:
                 action = 'DEFER_IF_PERMIT %s' % self.factory.settings.reject_msg
@@ -158,19 +192,92 @@ class BleyPolicy(PostfixPolicy):
                 self.factory.bad_cache[postfix_params['client_address']] = datetime.datetime.now()
             self.safe_execute(query, postfix_params)
 
-        else: # found to be clean
+        else:  # found to be clean
             check_results['DB'] = status[0]
             action = 'DUNNO'
-            query = "UPDATE bley_status SET last_action=NOW() WHERE ip=%(client_address)s AND sender=%(sender)s AND recipient=%(recipient)s"
+            query = "UPDATE bley_status SET last_action=%(now)s WHERE ip=%(client_address)s AND sender=%(sender)s AND recipient=%(recipient)s"
             self.safe_execute(query, postfix_params)
             self.factory.good_cache[postfix_params['client_address']] = datetime.datetime.now()
 
         if self.factory.settings.verbose:
-            self.factory.settings.logger('decided action=%s, checks: %s, postfix: %s\n' % (action, check_results, postfix_params))
+            logger.info('decided action=%s, checks: %s, postfix: %s' %
+                        (action, check_results, postfix_params))
         else:
-            self.factory.settings.logger('decided action=%s, from=%s, to=%s\n' % (action, postfix_params['sender'], postfix_params['recipient']))
+            logger.info('decided action=%s, from=%s, to=%s' %
+                        (action, postfix_params['sender'],
+                         postfix_params['recipient']))
+        self.factory.log_action(postfix_params, action, check_results)
         self.send_action(action)
 
+    def check_whitelist(self, email, whitelist):
+        '''Check the arg email against a whitelist
+        The whitelist consists of strings or compiled regular expressions
+        from the whitelist file
+        Return 1 if any of
+            email matches the entire entry
+            email matches the regular expression
+            email domain (or subdomain) matches the entry (which is a domain
+            name)
+            email user@ part matches an entry of the form user@
+        '''
+        for entry_wl in whitelist:
+            if isinstance(entry_wl, regexp_type):
+                if entry_wl.search(email) is not None:
+                    logger.info('whitelisted %s due to rule %s' %
+                                (email, entry_wl.pattern))
+                    return 1
+                else:
+                    # It's a regex, but it doesn't match
+                    # next whitelist item please
+                    continue
+            # whitelist item is a string
+            if entry_wl.endswith('@'):
+                # user@ (any domain) match
+                if email.startswith(entry_wl):
+                    logger.info('whitelisted %s due to rule %s' %
+                                (email, entry_wl))
+                    return 1
+                else:
+                    continue
+            if len(email) > len(entry_wl):
+                entry_wl_length = len(entry_wl) + 1
+                if ('.' + entry_wl) == email[-entry_wl_length:]:
+                    # subdomain match
+                    logger.info('whitelisted %s due to rule %s' %
+                                (email, entry_wl))
+                    return 1
+                elif ('@' + entry_wl) == email[-entry_wl_length:]:
+                    # @domain match
+                    logger.info('whitelisted %s due to rule %s' %
+                                (email, entry_wl))
+                    return 1
+            if entry_wl == email:
+                # Whole email match (for whitelist_recipients)
+                # Or domain match (for whitelist_clients)
+                logger.info('whitelisted %s due to rule %s' %
+                            (email, entry_wl))
+                return 1
+        return 0
+
+    def check_whitelist_ip(self, ipstr, whitelist_ip):
+        '''Check the arg ipstr against a whitelist
+        The whitelist consists of ip objects generated from the whitelist file
+        Return 1 if
+            ipstr is a valie ip address
+            -- AND --
+            ipstr matches one of the entries in the whitelist_ip list
+        '''
+        try:
+            ip = ipaddr.IPAddress(ipstr)
+        except (ValueError):
+            return 0
+        for net in whitelist_ip:
+            if ip in net:
+                logger.info('whitelisted %s because it is in subnet %s' %
+                            (str(ip), str(net)))
+                return 1
+        return 0
+
     def check_local_db(self, postfix_params):
         '''Check the sender for being in the local database.
 
@@ -182,7 +289,8 @@ class BleyPolicy(PostfixPolicy):
         @return: the result from SQL if any
         '''
 
-        query = """SELECT status,last_action,fail_count,sender,recipient FROM bley_status
+        query = """SELECT status,last_action,fail_count,sender,recipient
+                    FROM bley_status
                     WHERE ip=%(client_address)s
                     AND sender=%(sender)s AND recipient=%(recipient)s
                     ORDER BY status ASC
@@ -192,7 +300,7 @@ class BleyPolicy(PostfixPolicy):
             result = self.dbc.fetchone()
         except:
             result = None
-            self.factory.settings.logger('check_local_db failed. sending unknown.\n')
+            logger.info('check_local_db failed. sending unknown.')
         if not result:
             return -1
         else:
@@ -253,12 +361,14 @@ class BleyPolicy(PostfixPolicy):
         @return: twisted.names.client resolver
         '''
 
-        rip = reverse_ip(ip)
+        rip = bleyhelpers.reverse_ip(ip)
         lookup = '%s.%s' % (rip, lst)
         d = client.lookupAddress(lookup)
         return d
 
     def safe_execute(self, query, params=None):
+        if self.factory.settings.dbtype == 'sqlite3':
+            query = bleyhelpers.adapt_query_for_sqlite3(query)
         try:
             self.dbc.execute(query, params)
             self.db.commit()
@@ -268,11 +378,11 @@ class BleyPolicy(PostfixPolicy):
                 self.dbc.execute(query, params)
                 self.db.commit()
             else:
-                self.factory.settings.logger('Could not reconnect to the database, exiting.\n')
+                logger.info('Could not reconnect to the database, exiting.')
                 reactor.stop()
 
     def safe_reconnect(self):
-        self.factory.settings.logger('Reconnecting to the database after an error.\n')
+        logger.info('Reconnecting to the database after an error.')
         try:
             self.db.close()
         except:
@@ -287,7 +397,7 @@ class BleyPolicy(PostfixPolicy):
                 self.db = None
                 retries += 1
                 sleep(1)
-        
+
 
 class BleyPolicyFactory(Factory):
     protocol = BleyPolicy
@@ -296,3 +406,44 @@ class BleyPolicyFactory(Factory):
         self.settings = settings
         self.good_cache = {}
         self.bad_cache = {}
+        self.actionlog = []
+        self.exim_workaround = settings.exim_workaround
+        reactor.callLater(30 * 60, self.dump_log)
+        reactor.addSystemEventTrigger('before', 'shutdown', self.dump_log)
+
+    def log_action(self, postfix_params, action, check_results):
+        now = datetime.datetime.now()
+        action = action.split(' ')[0]
+        logline = {'time': str(now), 'ip': postfix_params['client_address'],
+                   'from': postfix_params['sender'],
+                   'to': postfix_params['recipient'],
+                   'action': action}
+        logline.update(check_results)
+        self.actionlog.append(logline)
+
+    def dump_log(self):
+        query = '''INSERT INTO bley_log (logtime, ip, sender, recipient,
+                action, check_dnswl, check_dnsbl, check_helo, check_dyn,
+                check_db, check_spf, check_s_eq_r, check_postmaster,
+                check_cache)
+                VALUES(%(time)s, %(ip)s, %(from)s, %(to)s, %(action)s,
+                %(DNSWL)s, %(DNSBL)s, %(HELO)s, %(DYN)s, %(DB)s,
+                %(SPF)s, %(S_EQ_R)s, %(WHITELISTED)s, %(CACHE)s)'''
+
+        if self.settings.dbtype == 'sqlite3':
+            query = bleyhelpers.adapt_query_for_sqlite3(query)
+
+        try:
+            db = self.settings.database.connect(**self.settings.dbsettings)
+            dbc = db.cursor()
+            i = len(self.actionlog)
+            while i:
+                logline = self.actionlog.pop(0)
+                dbc.execute(query, logline)
+                i -= 1
+            db.commit()
+            dbc.close()
+            db.close()
+        except self.settings.database.DatabaseError, e:
+            logger.warn('SQL error: %s' % e)
+        reactor.callLater(30 * 60, self.dump_log)
