@@ -1,4 +1,4 @@
-# Copyright (c) 2009 Evgeni Golov <evgeni.golov@uni-duesseldorf.de>
+# Copyright (c) 2009-2014 Evgeni Golov <evgeni@golov.de>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -27,10 +27,18 @@
 
 from twisted.protocols.basic import LineOnlyReceiver
 from twisted.internet.protocol import Factory
+from twisted.internet.interfaces import IHalfCloseableProtocol
+from zope.interface import implements
+import ipaddr
+
 
 class PostfixPolicy(LineOnlyReceiver):
     '''Basic implementation of a Postfix policy service.'''
 
+    implements(IHalfCloseableProtocol)
+
+    required_params = []
+
     def __init__(self):
         self.params = {}
         self.delimiter = '\n'
@@ -40,35 +48,56 @@ class PostfixPolicy(LineOnlyReceiver):
         line = line.strip().lower()
         if line == '':
             if len(self.params) > 0:
-                self.check_policy()
+                valid_request = True
+                for p in self.required_params:
+                    if p not in self.params:
+                        valid_request = False
+                        break
+                if valid_request:
+                    self.check_policy()
+                else:
+                    self.send_action('DUNNO')
             self.params = {}
         else:
             try:
                 (pkey, pval) = line.split('=', 1)
                 try:
-                    pval = pval.decode('utf-8', 'ignore').encode('us-ascii', 'ignore')
+                    pval = pval.decode('utf-8', 'ignore')
+                    pval = pval.encode('us-ascii', 'ignore')
                 except:
                     pass
+                if pkey == 'client_address':
+                    pval = ipaddr.IPAddress(pval).exploded
                 self.params[pkey] = pval
             except:
                 print 'Could not parse "%s"' % line
 
-    def check_policy (self):
+    def check_policy(self):
         '''Check the incoming mail based on our policy and tell Postfix
         about our decision.
- 
+
         You probably want to override this one with a function that does
         something useful.
         '''
         self.send_action('DUNNO')
 
-    def send_action (self, action='DUNNO'):
+    def send_action(self, action='DUNNO'):
         '''Send action back to Postfix.
- 
+
         @type action: string
         @param action: the action to be sent to Postfix (default: 'DUNNO')
         '''
         self.sendLine('action=%s\n' % action)
+        if self.factory.exim_workaround:
+            self.transport.loseConnection()
+
+    def readConnectionLost(self):
+        pass
+
+    def writeConnectionLost(self):
+        self.transport.loseConnection()
+
 
 class PostfixPolicyFactory(Factory):
     protocol = PostfixPolicy
+    exim_workaround = False
